const assert = @import("std").debug.assert;
const abs = @import("std").math.absCast; // builtin in 0.12 please

const intDist = @import("util.zig").intDist;

const map = @import("map.zig");
const Cardinal = map.Cardinal;
const Point = map.Point;
const TurningDirection = map.TurningDirection;


const algo = @import("algorithm.zig");
const robot = algo.robot;
const setWall = algo.setWall;

pub const Mouse = struct {
    position: map.Point,

    facing: Cardinal,

    pub fn spinTo(self: *Mouse, direction: Cardinal) void {
        if (self.facing == direction) {
            // return, we're done!
        }else if (
            (self.facing == .north and direction == .east) or
            (self.facing == .east and direction == .south) or
            (self.facing == .south and direction == .west) or
            (self.facing == .west and direction == .north)
        ) {
            //Rotate 90º Right
            robot.turn90(.right);
        }else if (
            (self.facing == .north and direction == .west) or
            (self.facing == .west and direction == .south) or
            (self.facing == .south and direction == .east) or
            (self.facing == .east and direction == .north)
        ) {
            //Rotate 90º Left
            robot.turn90(.left);
        }else if (
            (self.facing == .north and direction == .south) or
            (self.facing == .south and direction == .north) or
            (self.facing == .east and direction == .west) or
            (self.facing == .west and direction == .east)
        ) {
            // Rotate 180º
            robot.turn180();
        }

        self.facing = direction;
    }

    pub fn moveAdjacent (self: *Mouse, p: Point) void {
        // |self.position.x - p.x| + |self.position.y - p.y| == 1
        assert(intDist(self.position.x, p.x) + intDist(self.position.y, p.y) == 1);

        // The direction we need to move to get to this point
        var direction: Cardinal = undefined;
        // - 1 has a chance to underflow if it's 0.
        if (p.x + 1 == self.position.x) {
            direction = .west;
        }else if (p.y + 1 == self.position.y) {
            direction = .south;
        }else if (p.x > 0 and p.x - 1 == self.position.x) {
            direction = .east;
        }else if (p.y > 0 and p.y - 1 == self.position.y) {
            direction = .north;
        }

        self.spinTo(direction);

        // I haven't decided if this is going to be implicit 1 square or what
        robot.moveForward();

        // Update our position
        self.position = p;
    }

    // These functions return null if that direction is the back (where we don't have a sensor)
    // There might be a better way to do this, but eh
    fn readNorth(self: Mouse) ?bool {
        switch (self.facing) {
            .north => return robot.readFront(),
            .east => return robot.readLeft(),
            .south => return null,
            .west => return robot.readRight()
        }
    }
    fn readEast(self: Mouse) ?bool {
        switch (self.facing) {
            .north => return robot.readRight(),
            .east => return robot.readFront(),
            .south => return robot.readLeft(),
            .west => return null
        }
    }
    fn readSouth(self: Mouse) ?bool {
        switch (self.facing) {
            .north => return null,
            .east => return robot.readRight(),
            .south => return robot.readFront(),
            .west => return robot.readLeft()
        }
    }
    fn readWest(self: Mouse) ?bool {
        switch (self.facing) {
            .north => return robot.readLeft(),
            .east => return null,
            .south => return robot.readRight(),
            .west => return robot.readFront()
        }
    }

    pub fn updateWalls(self: Mouse) void {
        // TODO: Write to the simulator for cosmetics

        // I could probably come up with a read function that takes a direction and yada yada
        if (self.readNorth()) |isWall| {
            setWall(self.position, .north, isWall);
        }
        if (self.readEast()) |isWall| {
            setWall(self.position, .east, isWall);
        }
        if (self.readSouth()) |isWall| {
            setWall(self.position, .south, isWall);
        }
        if (self.readWest()) |isWall| {
            setWall(self.position, .west, isWall);
        }
    }
};

pub fn initialize() Mouse {
    return Mouse {
        .position = map.start,
        .facing = .north
    };
}
